@@ -1,6 +1,6 @@
 (***********************************************************************)
 (*                                                                     *)
-(*                           Objective Caml                            *)
+(*                                OCaml                                *)
 (*                                                                     *)
 (*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
 (*                                                                     *)
@@ -10,15 +10,22 @@
 (*                                                                     *)
 (***********************************************************************)
 
-(* $Id: lambda.ml 11113 2011-07-07 14:32:00Z maranget $ *)
-
 open Misc
 open Path
 open Asttypes
 
+type compile_time_constant =
+  | Big_endian
+  | Word_size
+  | Ostype_unix
+  | Ostype_win32
+  | Ostype_cygwin
+
 type primitive =
     Pidentity
   | Pignore
+  | Prevapply of Location.t
+  | Pdirapply of Location.t
     (* Globals *)
   | Pgetglobal of Ident.t
   | Psetglobal of Ident.t
@@ -84,6 +91,28 @@ type primitive =
   (* Operations on big arrays: (unsafe, #dimensions, kind, layout) *)
   | Pbigarrayref of bool * int * bigarray_kind * bigarray_layout
   | Pbigarrayset of bool * int * bigarray_kind * bigarray_layout
+  (* size of the nth dimension of a big array *)
+  | Pbigarraydim of int
+  (* load/set 16,32,64 bits from a string: (unsafe)*)
+  | Pstring_load_16 of bool
+  | Pstring_load_32 of bool
+  | Pstring_load_64 of bool
+  | Pstring_set_16 of bool
+  | Pstring_set_32 of bool
+  | Pstring_set_64 of bool
+  (* load/set 16,32,64 bits from a
+     (char, int8_unsigned_elt, c_layout) Bigarray.Array1.t : (unsafe) *)
+  | Pbigstring_load_16 of bool
+  | Pbigstring_load_32 of bool
+  | Pbigstring_load_64 of bool
+  | Pbigstring_set_16 of bool
+  | Pbigstring_set_32 of bool
+  | Pbigstring_set_64 of bool
+  (* Compile time constants *)
+  | Pctconst of compile_time_constant
+  (* byte swap *)
+  | Pbswap16
+  | Pbbswap of boxed_integer
 
 and comparison =
     Ceq | Cneq | Clt | Cgt | Cle | Cge
@@ -240,7 +269,7 @@ let name_lambda_list args fn =
       Llet(Strict, id, arg, name_list (Lvar id :: names) rem) in
   name_list [] args
 
-let rec iter f = function
+let iter f = function
     Lvar _
   | Lconst _ -> ()
   | Lapply(fn, args, _) ->
@@ -446,6 +475,7 @@ let transl_location loc =
 (*>JOCAML*)
 let may_raise = function
   | Pccall desc -> desc.Primitive.prim_alloc
+  | Prevapply _|Pdirapply _ (* call function, anything can happen *)
   | Praise 
   | Pstringrefs | Pstringsets
   | Parrayrefs _
@@ -456,6 +486,11 @@ let may_raise = function
   | Pbigarrayref _
   | Pbigarrayset _
   | Plazyforce
+(* New in 4.01, err on the safe side... *)
+  | Pbigarraydim _|Pstring_load_16 _|Pstring_load_32 _|Pstring_load_64 _|
+    Pstring_set_16 _|Pstring_set_32 _|Pstring_set_64 _|Pbigstring_load_16 _|
+    Pbigstring_load_32 _|Pbigstring_load_64 _|Pbigstring_set_16 _|
+    Pbigstring_set_32 _|Pbigstring_set_64 _|Pctconst _|Pbswap16|Pbbswap _
     -> true
   | Pidentity
   | Pignore
