@@ -1,6 +1,6 @@
 (***********************************************************************)
 (*                                                                     *)
-(*                           Objective Caml                            *)
+(*                                OCaml                                *)
 (*                                                                     *)
 (*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
 (*                                                                     *)
@@ -10,28 +10,25 @@
 (*                                                                     *)
 (***********************************************************************)
 
-(* $Id: translcore.mli 9081 2008-10-14 07:37:28Z maranget $ *)
-
 (* Translation from typed abstract syntax to lambda terms,
    for the core language *)
 
 open Asttypes
-open Types
 open Typedtree
 open Lambda
 
 val name_pattern: string -> (pattern * 'a) list -> Ident.t
 
 val transl_exp: expression -> lambda
-val transl_apply: lambda -> (expression option * optional) list
+val transl_apply: lambda -> (label * expression option * optional) list
                   -> Location.t -> lambda
 val transl_let:
       rec_flag -> (pattern * expression) list -> lambda -> lambda
 (*> JOCAML *)
 val transl_def: joinautomaton list -> lambda -> lambda
-val transl_loc: joinlocation list -> lambda -> lambda    
 (*< JOCAML *)
-val transl_primitive: Primitive.description -> lambda
+val transl_primitive: Location.t -> Primitive.description -> lambda
+
 val transl_exception:
       Ident.t -> Path.t option -> exception_declaration -> lambda
 
@@ -41,6 +38,7 @@ type error =
     Illegal_letrec_pat
   | Illegal_letrec_expr
   | Free_super_var
+  | Unknown_builtin_primitive of string
 
 exception Error of Location.t * error
 
