@@ -10,7 +10,7 @@
 (*                                                                     *)
 (***********************************************************************)
 
-(* $Id: transljoin.ml 10506 2010-06-04 19:16:52Z maranget $ *)
+(* $Id: transljoin.ml 14456 2014-03-13 12:44:09Z maranget $ *)
 
 open Misc
 open Longident
@@ -18,7 +18,9 @@ open Primitive
 open Types
 open Asttypes
 open Typedtree
-open Env
+(*>JOCAML *)
+open Auto
+(*<JOCAML *)
 open Lambda
 
 (* DEBUG stuff *)
@@ -221,10 +223,10 @@ let principal p = match Typejoin.get_replies p with
 let rec is_principal id p = match p.exp_desc with
 |  Texp_asyncsend (_,_) | Texp_null 
  -> false
-| Texp_reply (_, kont) -> kont=id
+| Texp_reply (_, kont,_) -> kont=id
 | Texp_par (p1, p2) ->
    is_principal id p1 || is_principal id p2
-| Texp_let (_,_,p) | Texp_def (_,p) | Texp_loc (_,p)
+| Texp_let (_,_,p) | Texp_def (_,p)
 | Texp_sequence (_,p) | Texp_when (_,p) -> 
    is_principal id p
 | Texp_match (_,(_,p)::cls,_) ->
@@ -233,7 +235,7 @@ let rec is_principal id p = match p.exp_desc with
 | Texp_ifthenelse (_,pifso, Some pifno) ->
    is_principal id pifso && is_principal id pifno
 | Texp_ifthenelse (_,_,None) -> false
-| Texp_for (_,_,_,_,_) -> false
+| Texp_for (_,_,_,_,_,_) -> false
 | _ -> assert false
 
 (*
@@ -263,11 +265,11 @@ let simple_prim = ref ((fun p -> assert false) : Primitive.description -> bool)
 
 let rec simple_pat p = match p.pat_desc with
 | Tpat_any | Tpat_var _ -> true
-| Tpat_alias (p,_)|Tpat_lazy p -> simple_pat p
+| Tpat_alias (p,_,_)|Tpat_lazy p -> simple_pat p
 | Tpat_tuple ps -> List.for_all simple_pat ps
-| Tpat_record lps ->  List.for_all (fun (_,p) -> simple_pat p) lps
+| Tpat_record (lps,_) ->  List.for_all (fun (_,_,p) -> simple_pat p) lps
 | Tpat_or (p1,p2,_) -> simple_pat p1 && simple_pat p2
-| Tpat_constant _|Tpat_construct (_,_)|Tpat_variant (_,_,_)
+| Tpat_constant _|Tpat_construct _|Tpat_variant _
 | Tpat_array _ -> false
 
 let rec simple_exp e = match e.exp_desc with
@@ -283,38 +285,38 @@ let rec simple_exp e = match e.exp_desc with
 | Texp_match (_, _, Partial) -> false
 | Texp_ifthenelse (e, eifso, eo) ->
    simple_exp e && simple_exp eifso && simple_exp_option eo
-| Texp_def (_,e)|Texp_loc(_,e) -> simple_exp e
+| Texp_def (_,e) -> simple_exp e
 (* Simple simple expressions *)
-| Texp_ident _ | Texp_constant _ | Texp_function (_,_)
+| Texp_ident _ | Texp_constant _ | Texp_function _
 | Texp_variant (_,None) 
-| Texp_instvar (_,_) | Texp_setinstvar (_, _, _) | Texp_spawn (_)
+| Texp_instvar _ | Texp_setinstvar _ | Texp_spawn (_)
  -> true
 (* Recursion *)
-| Texp_construct (_,es) | Texp_tuple (es) | Texp_array (es)
+| Texp_construct (_,_,es,_) | Texp_tuple (es) | Texp_array (es)
  -> List.for_all simple_exp es
-| Texp_variant (_, Some e) | Texp_field (e,_)
+| Texp_variant (_, Some e) | Texp_field (e,_,_)
    -> simple_exp e
-| Texp_setfield (e1,_,e2) -> simple_exp e1 && simple_exp e2
-| Texp_apply ({exp_desc=Texp_ident (_, {val_kind=Val_prim p})}, args) ->
+| Texp_setfield (e1,_,_,e2) -> simple_exp e1 && simple_exp e2
+| Texp_apply ({exp_desc=Texp_ident (_,_,{val_kind=Val_prim p})}, args) ->
    List.length args < p.prim_arity || (* will be compiled as function *)
    (!simple_prim p &&
-   List.for_all (fun (eo,_) -> simple_exp_option eo) args)
+   List.for_all (fun (_,eo,_) -> simple_exp_option eo) args)
 | Texp_apply (_,_) -> false
-| Texp_for (_,e1,e2,_,e3) ->
+| Texp_for (_,_,e1,e2,_,e3) ->
    simple_exp e1 && simple_exp e2 && simple_exp e3
 | Texp_record (les,eo) ->
-   List.for_all (fun (_,e) -> simple_exp e) les &&
+   List.for_all (fun (_,_,e) -> simple_exp e) les &&
    simple_exp_option eo
 (* Asserts are special *)
 | Texp_assert e -> !Clflags.noassert || simple_exp e
 | Texp_assertfalse -> !Clflags.noassert
 (* Who knows ? *)
-| Texp_letmodule (_,_,_) | Texp_override (_,_) | Texp_lazy (_)
-| Texp_send (_,_) | Texp_while (_,_) | Texp_new (_,_) | Texp_try (_,_)
-| Texp_object (_, _, _) | Texp_pack _
+| Texp_letmodule _ | Texp_override (_,_) | Texp_lazy (_)
+| Texp_send _ | Texp_while (_,_) | Texp_new _ | Texp_try (_,_)
+| Texp_object _| Texp_pack _
   -> false
 (* Process constructs are not errors *)
-| Texp_reply (_, _)|Texp_par (_, _)|Texp_asyncsend (_, _)
+| Texp_reply _|Texp_par (_, _)|Texp_asyncsend (_, _)
 | Texp_null
  -> assert false
 
@@ -337,22 +339,22 @@ and simple_proc p = match p.exp_desc with
    simple_exp e && simple_proc pifso && simple_proc pifno
 | Texp_ifthenelse (e, pifso, None) ->
    simple_exp e && simple_proc pifso    
-| Texp_def (_,p)|Texp_loc(_,p) -> simple_proc p
-| Texp_for (_,e1,e2,_,_body) -> (* _body is compiled so a not to fail *)
+| Texp_def (_,p) -> simple_proc p
+| Texp_for (_,_,e1,e2,_,_body) -> (* _body is compiled so a not to fail *)
    simple_exp e1 && simple_exp e2
 (* Process constructs *)
-| Texp_reply (e, _) -> simple_exp e
+| Texp_reply (e, _, _) -> simple_exp e
 | Texp_par (p1, p2) -> simple_proc p1 || simple_proc p2
 | Texp_asyncsend (e1, e2) -> simple_exp e1 && simple_exp e2
 | Texp_null -> true
 (* Plain expressions no longer are errors *)
-| Texp_spawn _|Texp_object (_, _, _)|Texp_lazy _|Texp_assert _|
-  Texp_letmodule (_, _, _)|Texp_override (_, _)|Texp_setinstvar (_, _, _)|
-  Texp_instvar (_, _)|Texp_new (_, _)|Texp_send (_, _)|
-  Texp_while (_, _)|Texp_array _|
-  Texp_setfield (_, _, _)|Texp_field (_, _)|Texp_record (_, _)|
-  Texp_variant (_, _)|Texp_construct (_, _)|Texp_tuple _|Texp_try (_, _)|
-  Texp_apply (_, _)|Texp_function (_, _)|Texp_constant _|Texp_ident (_, _)|
+| Texp_spawn _|Texp_object _|Texp_lazy _|Texp_assert _|
+  Texp_letmodule _|Texp_override _|Texp_setinstvar _|
+  Texp_instvar _|Texp_new _|Texp_send _|
+  Texp_while _|Texp_array _|
+  Texp_setfield _|Texp_field _|Texp_record _|
+  Texp_variant _|Texp_construct _|Texp_tuple _|Texp_try _|
+  Texp_apply _|Texp_function _|Texp_constant _|Texp_ident _|
   Texp_assertfalse|Texp_pack _
   -> assert false
 
@@ -421,14 +423,130 @@ let as_procs sync e =
 let rec get_chan_rec id = function
  | [] -> raise Not_found
  | (oid,x)::rem ->
-     if id = oid then x else get_chan_rec id rem
+     if Ident.same id oid then x else get_chan_rec id rem
 
-let dump_idx fp (id, _) = fprintf fp "%s" (Ident.unique_name id)
+open Printf
 
+let dump_chan fp chan =
+  fprintf fp "{id=%s, sync=%b}"
+    (Ident.unique_name chan.jchannel_ident)
+    chan.jchannel_sync
+
+let compile_auto a =
+(*
+  eprintf "AUTO:" ;
+  List.iter (fun (_,chan) -> eprintf " %a" dump_chan chan) a.jauto_names ;
+  eprintf "\n%!" ;
+*)
+  let get_chan tag id =
+    try get_chan_rec id a.jauto_names
+    with Not_found ->
+      fatal_error
+        (Printf.sprintf "Transljoin.compile_auto [%s]: %s" tag
+           (Ident.unique_name id)) in
+          
+  let (disps, reacs), new_names = 
+    Joinmatching.compile a.jauto_loc
+      (List.map
+         (fun cl ->
+           let jpats,g = cl.jclause_desc in
+           cl.jclause_loc,jpats,g)
+         a.jauto_desc) in
+(* Allocate new names for dispatchers and guarded processes *)
+  let disps =  List.map (fun disp -> Ident.create "#d#", disp) disps
+  and reacs,fwds = match reacs with
+  | [((_,[_],_),_) as reac] -> (* Case of (single) forwarder *)
+      [],[Ident.create "#f", reac]
+  | _ ->
+      List.map (fun reac -> Ident.create "#g#", reac) reacs, [] in
+(* Automaton names *)
+   let name = Ident.create "#auto#"
+   and name_wrapped = Ident.create "#wrapped#" in
+(* Collect forwarders *)
+  let fwd_names =
+    List.map
+      (fun (d_id,(id,_,_)) ->
+        let chan = get_chan "dispatchers" id in
+        id,(chan,(Alone d_id)))
+      disps in
+  let fwd_names =
+    List.fold_right
+      (fun fwd r ->
+        let g_id,(_old,(patss,_gd)) = fwd in
+         List.fold_right
+          (fun pats r ->
+            List.fold_right
+              (fun pat r ->
+                let jid,_ = pat.jpat_desc in
+                let id = jid.jident_desc in
+                let chan = get_chan "forwarders" id in
+                (id,(chan,Alone g_id))::r)
+              pats r)
+          patss r)
+      fwds fwd_names in
+(* Collect channels *)
+  let is_fwd id =
+    try ignore(get_chan_rec id fwd_names) ; true
+    with Not_found -> false in
+
+(* First orignal names unsplit by compilation, and not a forwarder *)
+  let all_names,nchans =
+    List.fold_right
+      (fun (id,chan) (chans,nchans as k) ->
+        if is_fwd id then k
+        else
+          (id,(chan,Chan (name,nchans)))::chans,nchans+1)
+      a.jauto_names (fwd_names,0) in
+(* Then new names, introduced by pattern matching compilation *)
+  let all_names,nchans =
+     List.fold_right
+      (fun (id,ids) k ->
+        let chan = get_chan "old_name" id in
+        List.fold_right
+          (fun id (chans,nchans) ->
+            let chan = { chan with jchannel_ident = id} in
+            (id,(chan,Chan (name,nchans)))::chans,nchans+1)
+          ids k)
+      new_names (all_names,nchans) in
+(* Original names, with information *)
+  let is_orig id =
+     try ignore(get_chan_rec id a.jauto_names) ; true
+    with Not_found -> false in
+  let original = List.filter (fun (id,_) -> is_orig id) all_names in
+(* Precompile dispatchers *)
+  let disps =
+    let find_channel id =
+      try get_chan_rec id all_names
+      with Not_found ->
+        fatal_error
+        (Printf.sprintf "find_channel: %s"
+           (Ident.unique_name id)) in
+    List.map
+      (fun disp ->
+        let d_id,(chan_id, cls, par) = disp in
+        let chan = find_channel chan_id
+        and cls = List.map (fun (p,id) -> p,find_channel id) cls in
+        d_id,chan,cls,par)
+      disps in
+(* Precompile reactions, little to do: flatten *)
+  let precomp_reac (id,((_loc,pat,e),(pats,bv))) = id,pat,pats,bv,e in
+  let fwds = List.map precomp_reac fwds
+  and reacs = List.map precomp_reac reacs in
+(* Gather everything... *)
+  {
+    cauto_name = name,name_wrapped;
+    cauto_channels = all_names;
+    cauto_nchans = nchans;
+    cauto_original = original;
+    cauto_loc = a.jauto_loc;
+    cauto_dispatchers = disps;
+    cauto_forwarders = fwds;
+    cauto_reactions = reacs;
+  }
 let get_num msg names id =
  try
-   let {jchannel_id=x} = get_chan_rec id names in
-   match x with
+   let _,opt = get_chan_rec id names in
+   match opt with
    | Chan (_,num) -> num
    | _ ->
      fatal_error
@@ -441,8 +559,8 @@ let get_num msg names id =
 
 let get_chan msg names id =
   try
-   let {jchannel_id=x} = get_chan_rec id names in
-   x
+   let _,opt = get_chan_rec id names in
+   opt
  with
  | Not_found ->
      fatal_error
@@ -464,17 +582,18 @@ let rec principal_param ipri params nums = match params, nums with
 let names_block nchans names =
   let t = Array.create nchans "" in
   List.iter
-    (fun (id, {jchannel_id=x}) ->match x with
-    | Chan (_,i) -> t.(i) <- Ident.unique_name id
-    | _ -> ())
+    (fun (id, (_,opt)) ->
+      match opt with
+      | Chan (_,i) -> t.(i) <- Ident.unique_name id
+      | _ -> ())
     names ;
   Lconst
     (Const_block
        (0, Array.fold_right (fun s r -> Const_base (Const_string s)::r) t []))
 
 let create_auto
-   { jauto_name=(auto_name, wrapped_name);
-     jauto_names = names ; jauto_nchans=nchans ; } k =
+   { cauto_name=(auto_name, wrapped_name);
+     cauto_channels = names ; cauto_nchans=nchans ; } k =
   if nchans > 0 then
     Llet
       (Strict, auto_name,
@@ -485,15 +604,14 @@ let create_auto
          (Strict, wrapped_name, wrap_automaton auto_name Location.none, k))
   else k
 
-let create_channels {jauto_name=(raw_name, name) ; jauto_names=names} k =
+let create_channels {cauto_name=(raw_name, name) ; cauto_channels=names} k =
   List.fold_right
-    (fun (id,jc) k ->
-      let {jchannel_sync=sync ; jchannel_id=x} = jc in
-      match x with
+    (fun (id,(jc,opt)) k ->
+      match opt with
       | Chan (_,num) ->
 	  Llet
 	    (StrictOpt, id,
-	     (if sync then
+	     (if jc.jchannel_sync then
                create_sync
 	     else
                create_async) name num Location.none,
@@ -543,10 +661,10 @@ let create_forwarders autos dispss fwdss r =
     List.fold_right
       (fun auto r ->
         List.fold_right
-          (fun (_,chan) r -> match chan.jchannel_id with
+          (fun (_,(chan,opt)) r -> match opt with
           | Alone g -> make_g "fwd" chan g::r
           | Chan (_,_) -> r)
-          auto.jauto_names r)
+          auto.cauto_channels r)
       autos id2g in
 (* patch forwarder data structure *)
    let r =
@@ -707,13 +825,13 @@ let lapply (f,args) = Lapply (f,args,Location.none)
 (* gs is a list of compiled guarded processes *)
 
 let create_table auto gs r =
-  let n_chans = auto.jauto_nchans in
+  let n_chans = auto.cauto_nchans in
   if n_chans =0 then r
   else
-    let name,_ = auto.jauto_name (* wrapped name of automaton *)
-    and names = auto.jauto_names in (* all channels *)
+    let name,_ = auto.cauto_name       (* wrapped name of automaton *)
+    and names = auto.cauto_channels in (* all channels *)
 
-    let rec do_guard (Reac reac) (_, sync, _) k =
+    let do_guard reac (_, sync, _) k =
       let (g, _, actual, _, _) = reac in
 
       let create_reaction jpats r =
@@ -773,7 +891,7 @@ let create_table auto gs r =
       let pats = explode actual in
       List.fold_right create_reaction pats k in
 
-    let _, reacs, _ = auto.jauto_desc in
+    let reacs = auto.cauto_reactions in
     Lsequence
       (patch_table name
          (List.fold_right2 do_guard reacs gs []) Location.none,
