@@ -10,7 +10,7 @@
 (*                                                                     *)
 (***********************************************************************)
 
-(* $Id: transljoin.mli 9082 2008-10-14 08:08:38Z maranget $ *)
+(* $Id: transljoin.mli 12959 2012-09-27 13:12:51Z maranget $ *)
 
 
 (**********************************************)
@@ -72,25 +72,24 @@ val principal : Typedtree.expression -> Ident.t option
 
 (* Building definitions and locations *)
 
-val create_auto : 
-  'a Typedtree.joinautomaton_gen ->
-    Lambda.lambda -> Lambda.lambda
+val compile_auto : Typedtree.joinautomaton ->  Auto.t
 
-val create_channels :
-  'a Typedtree.joinautomaton_gen -> Lambda.lambda -> Lambda.lambda
+val create_auto : Auto.t -> Lambda.lambda -> Lambda.lambda
+
+val create_channels :  Auto.t -> Lambda.lambda -> Lambda.lambda
 
 val create_dispatchers :
      (Ident.t * Typedtree.joinchannel * Lambda.lambda) list ->
        Lambda.lambda -> Lambda.lambda
 
 val create_forwarders :
-    'a Typedtree.joinautomaton_gen list ->
+    Auto.t list ->
       (Ident.t * Typedtree.joinchannel * Lambda.lambda) list list ->
          (Ident.t * Lambda.lambda) list list ->
            Lambda.lambda -> Lambda.lambda
 
 val create_table:
-    Typedtree.joinautomaton ->
+    Auto.t ->
       (Ident.t * Ident.t option * 'a) list -> Lambda.lambda ->
       Lambda.lambda
             
