@@ -1,6 +1,6 @@
 /***********************************************************************/
 /*                                                                     */
-/*                           Objective Caml                            */
+/*                                OCaml                                */
 /*                                                                     */
 /*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         */
 /*                                                                     */
@@ -11,8 +11,6 @@
 /*                                                                     */
 /***********************************************************************/
 
-/* $Id: intext.h 7192 2005-10-27 09:14:16Z maranget $ */
-
 /* Structured input/output */
 
 #ifndef CAML_INTEXT_H
@@ -86,6 +84,10 @@ void caml_output_val (struct channel * chan, value v, value flags);
 
 /* </private> */
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 CAMLextern void caml_output_value_to_malloc(value v, value flags,
                                             /*out*/ char ** buf,
                                             /*out*/ intnat * len);
@@ -110,7 +112,7 @@ void caml_intern_cleanup(void) ;
 /* </private> */
 
 CAMLextern value caml_input_val_from_string (value str, intnat ofs);
-  /* Read a structured value from the Caml string [str], starting
+  /* Read a structured value from the OCaml string [str], starting
      at offset [ofs]. */
 CAMLextern value caml_input_value_from_malloc(char * data, intnat ofs);
   /* Read a structured value from a malloced buffer.  [data] points
@@ -157,23 +159,27 @@ CAMLextern void caml_deserialize_error(char * msg);
 /* <private> */
 
 /* Auxiliary stuff for sending code pointers */
-unsigned char * caml_code_checksum (void);
 
-#ifndef NATIVE_CODE
-#include "fix_code.h"
-#define caml_code_area_start ((char *) caml_start_code)
-#define caml_code_area_end ((char *) caml_start_code + caml_code_size)
-#else
-extern char * caml_code_area_start, * caml_code_area_end;
-#endif
+struct code_fragment {
+  char * code_start;
+  char * code_end;
+  unsigned char digest[16];
+  char digest_computed;
+};
+
+struct ext_table caml_code_fragments_table;
 
 /* </private> */
 
 /* >JOCAML */
 CAMLextern void extern_invalid_argument(char *msg) ;
-CAMLextern code_t caml_get_saved_code(int idx) ;
+CAMLextern char *caml_get_saved_code(int idx) ;
 CAMLextern value caml_get_saved_value(int idx) ;
-/*<JOCAML*/
 /* <JOCAML */
 
+
+#ifdef __cplusplus
+}
+#endif
+
 #endif /* CAML_INTEXT_H */
