@@ -1,6 +1,6 @@
 /***********************************************************************/
 /*                                                                     */
-/*                           Objective Caml                            */
+/*                                OCaml                                */
 /*                                                                     */
 /*             Damien Doligez, projet Para, INRIA Rocquencourt         */
 /*                                                                     */
@@ -11,8 +11,6 @@
 /*                                                                     */
 /***********************************************************************/
 
-/* $Id: minor_gc.h 9081 2008-10-14 07:37:28Z maranget $ */
-
 #ifndef CAML_MINOR_GC_H
 #define CAML_MINOR_GC_H
 
@@ -39,7 +37,7 @@ CAMLextern struct caml_ref_table caml_ref_table, caml_weak_ref_table;
   (Assert (Is_block (val)), \
    (addr)(val) < (addr)caml_young_end && (addr)(val) > (addr)caml_young_start)
 
-extern void caml_set_minor_heap_size (asize_t);
+extern void caml_set_minor_heap_size (asize_t); /* size in bytes */
 extern void caml_empty_minor_heap (void);
 CAMLextern void caml_minor_collection (void);
 CAMLextern void garbage_collection (void); /* def in asmrun/signals.c */
