@@ -10,10 +10,10 @@
 (*                                                                     *)
 (***********************************************************************)
 
-(* $Id: join_message.ml 8138 2007-03-29 14:36:25Z maranget $ *)
+(* $Id: join_message.ml 10588 2010-06-17 16:37:18Z maranget $ *)
 
 open Join_types
-(*DEBUG*)open Printf
+open Printf
 (*DEBUG*)open Join_debug
 
 (*****************************)
