@@ -10,7 +10,7 @@
 (*                                                                     *)
 (***********************************************************************)
 
-(* $Id: join_misc.ml 8337 2007-06-12 13:00:25Z maranget $ *)
+(* $Id: join_misc.ml 12965 2012-09-27 16:06:22Z maranget $ *)
 
 (*DEBUG*)open Join_debug
 
@@ -129,7 +129,7 @@ let decr c =
   Mutex.unlock c.cmutex
 
 
-let rec hard_wait_zero c = match c.cval with
+let hard_wait_zero c = match c.cval with
 | 0 -> ()
 | _ -> Condition.wait c.ccond c.cmutex
 
