@@ -1,3 +1,15 @@
+(***********************************************************************)
+(*                                                                     *)
+(*                                OCaml                                *)
+(*                                                                     *)
+(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
+(*                                                                     *)
+(*  Copyright 2000 Institut National de Recherche en Informatique et   *)
+(*  en Automatique.  All rights reserved.  This file is distributed    *)
+(*  under the terms of the Q Public License version 1.0.               *)
+(*                                                                     *)
+(***********************************************************************)
+
 (* Test the types nativeint, int32, int64 *)
 
 open Printf
@@ -28,30 +40,30 @@ let test test_number answer correct_answer =
 module type TESTSIG = sig
   type t
   module Ops : sig
-    val neg: t -> t 
-    val add: t -> t -> t 
-    val sub: t -> t -> t 
-    val mul: t -> t -> t 
-    val div: t -> t -> t 
-    val rem: t -> t -> t 
-    val logand: t -> t -> t 
-    val logor: t -> t -> t 
-    val logxor: t -> t -> t 
-    val shift_left: t -> int -> t 
-    val shift_right: t -> int -> t 
-    val shift_right_logical: t -> int -> t 
-    val of_int: int -> t 
-    val to_int: t -> int 
-    val of_float: float -> t 
+    val neg: t -> t
+    val add: t -> t -> t
+    val sub: t -> t -> t
+    val mul: t -> t -> t
+    val div: t -> t -> t
+    val rem: t -> t -> t
+    val logand: t -> t -> t
+    val logor: t -> t -> t
+    val logxor: t -> t -> t
+    val shift_left: t -> int -> t
+    val shift_right: t -> int -> t
+    val shift_right_logical: t -> int -> t
+    val of_int: int -> t
+    val to_int: t -> int
+    val of_float: float -> t
     val to_float: t -> float
     val zero: t
     val one: t
     val minus_one: t
     val min_int: t
     val max_int: t
-    val format : string -> t -> string 
+    val format : string -> t -> string
     val to_string: t -> string
-    val of_string: string -> t 
+    val of_string: string -> t
   end
   val testcomp: t -> t -> bool*bool*bool*bool*bool*bool*int
   val skip_float_tests: bool
@@ -166,6 +178,7 @@ struct
        9, 127531236, -365;
        10, 1234567, 12345678;
        11, 1234567, -12345678];
+    test 12 (div min_int (of_int (-1))) min_int;
 
     testing_function "mod";
     List.iter
@@ -181,6 +194,7 @@ struct
        9, 127531236, -365;
        10, 1234567, 12345678;
        11, 1234567, -12345678];
+    test 12 (rem min_int (of_int (-1))) (of_int 0);
 
     testing_function "and";
     List.iter
@@ -345,7 +359,7 @@ struct
     test 5 (add (of_int (-123)) (of_int 456)) (of_int 333);
     test 6 (add (of_int 123) (of_int (-456))) (of_int (-333));
     test 7 (add (of_int (-123)) (of_int (-456))) (of_int (-579));
-    test 8 (add (of_string "0x1234567812345678") 
+    test 8 (add (of_string "0x1234567812345678")
                 (of_string "0x9ABCDEF09ABCDEF"))
            (of_string "0x1be024671be02467");
     test 9 (add max_int max_int) (of_int (-2));
@@ -362,7 +376,7 @@ struct
     test 5 (sub (of_int (-123)) (of_int 456)) (of_int (-579));
     test 6 (sub (of_int 123) (of_int (-456))) (of_int 579);
     test 7 (sub (of_int (-123)) (of_int (-456))) (of_int 333);
-    test 8 (sub (of_string "0x1234567812345678") 
+    test 8 (sub (of_string "0x1234567812345678")
                 (of_string "0x9ABCDEF09ABCDEF"))
            (of_string "0x888888908888889");
     test 9 (sub max_int min_int) minus_one;
@@ -400,6 +414,7 @@ struct
        9, 127531236, -365;
        10, 1234567, 12345678;
        11, 1234567, -12345678];
+    test 12 (div min_int (of_int (-1))) min_int;
 
     testing_function "mod";
     List.iter
@@ -415,6 +430,7 @@ struct
        9, 127531236, -365;
        10, 1234567, 12345678;
        11, 1234567, -12345678];
+    test 12 (rem min_int (of_int (-1))) (of_int 0);
 
     testing_function "and";
     List.iter
@@ -524,7 +540,7 @@ let _ =
   begin match Sys.word_size with
     32 ->
       let module C =
-        Test32(struct type t = nativeint 
+        Test32(struct type t = nativeint
                       module Ops = Nativeint
                       let testcomp = testcomp_nativeint
                       let skip_float_tests = true end)
@@ -533,7 +549,7 @@ let _ =
       let module C =
         Test64(struct type t = nativeint
                       module Ops = Nativeint
-                      let testcomp = testcomp_nativeint 
+                      let testcomp = testcomp_nativeint
                       let skip_float_tests = true end)
       in ()
   | _ ->
@@ -549,7 +565,7 @@ let _ =
   test 3 (Nativeint.to_int32 (Nativeint.of_string "0x123456789ABCDEF0"))
          (Int32.of_string "0x9ABCDEF0")
   else
-  test 3 0 0; (* placeholder to have the same output on both 32-bit and 64-bit *)
+  test 3 0 0; (* placeholder to have the same output on 32-bit and 64-bit *)
   testing_function "int64 of/to int32";
   test 1 (Int64.of_int32 (Int32.of_string "-0x12345678"))
          (Int64.of_string "-0x12345678");
