@@ -1,6 +1,6 @@
 (***********************************************************************)
 (*                                                                     *)
-(*                           Objective Caml                            *)
+(*                                OCaml                                *)
 (*                                                                     *)
 (*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
 (*                                                                     *)
@@ -10,8 +10,6 @@
 (*                                                                     *)
 (***********************************************************************)
 
-(* $Id: path.mli 7205 2005-10-27 16:29:22Z maranget $ *)
-
 (* Access paths *)
 
 type t =
@@ -25,7 +23,10 @@ val binding_time: t -> int
 
 val nopos: int
 
-val name: t -> string
+val name: ?paren:(string -> bool) -> t -> string
+    (* [paren] tells whether a path suffix needs parentheses *)
     (* Return the print name, e.g., "F(X).M.foo" *)
 
 val head: t -> Ident.t
+
+val last: t -> string
