@@ -1,6 +1,6 @@
 (***********************************************************************)
 (*                                                                     *)
-(*                           Objective Caml                            *)
+(*                                OCaml                                *)
 (*                                                                     *)
 (*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
 (*                                                                     *)
@@ -10,8 +10,6 @@
 (*                                                                     *)
 (***********************************************************************)
 
-(* $Id: predef.mli 8126 2007-03-26 15:38:32Z maranget $ *)
-
 (* Predefined type constructors (with special typing rules in typecore) *)
 
 open Types
@@ -66,3 +64,4 @@ val build_initial_env:
 (* To initialize linker tables *)
 
 val builtin_values: (string * Ident.t) list
+val builtin_idents: (string * Ident.t) list
