@@ -1,6 +1,6 @@
 (***********************************************************************)
 (*                                                                     *)
-(*                           Objective Caml                            *)
+(*                                OCaml                                *)
 (*                                                                     *)
 (*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
 (*                                                                     *)
@@ -10,40 +10,44 @@
 (*                                                                     *)
 (***********************************************************************)
 
-(* $Id: typedecl.mli 11113 2011-07-07 14:32:00Z maranget $ *)
-
 (* Typing of type definitions and primitive definitions *)
 
+open Asttypes
 open Types
 open Format
 
 val transl_type_decl:
-    Env.t -> (string * Parsetree.type_declaration) list ->
-                                  (Ident.t * type_declaration) list * Env.t
+    Env.t -> (string loc * Parsetree.type_declaration) list ->
+    (Ident.t * string Asttypes.loc * Typedtree.type_declaration) list * Env.t
+
 val transl_exception:
-    Env.t -> Parsetree.exception_declaration -> exception_declaration
+    Env.t -> Location.t ->
+    Parsetree.exception_declaration -> Typedtree.exception_declaration
 
 val transl_exn_rebind:
     Env.t -> Location.t -> Longident.t -> Path.t * exception_declaration
 
 (*>JOCAML*)
 val transl_exn_global:
-        Env.t -> Location.t -> Longident.t -> Path.t
+    Env.t -> Location.t -> Longident.t -> Path.t
 (*<JOCAML*)
 
 val transl_value_decl:
-    Env.t -> Parsetree.value_description -> value_description
+    Env.t -> Location.t ->
+    Parsetree.value_description -> Typedtree.value_description
 
 val transl_with_constraint:
-    Env.t -> Ident.t -> Path.t option -> type_declaration ->
-    Parsetree.type_declaration -> type_declaration
+    Env.t -> Ident.t -> Path.t option -> Types.type_declaration ->
+    Parsetree.type_declaration -> Typedtree.type_declaration
 
 val abstract_type_decl: int -> type_declaration
 val approx_type_decl:
-    Env.t -> (string * Parsetree.type_declaration) list ->
+    Env.t -> (string loc * Parsetree.type_declaration) list ->
                                   (Ident.t * type_declaration) list
 val check_recmod_typedecl:
     Env.t -> Location.t -> Ident.t list -> Path.t -> type_declaration -> unit
+val check_coherence:
+    Env.t -> Location.t -> Ident.t -> type_declaration -> unit
 
 (* for fixed types *)
 val is_fixed_type : Parsetree.type_declaration -> bool
@@ -51,10 +55,11 @@ val is_fixed_type : Parsetree.type_declaration -> bool
 (* for typeclass.ml *)
 val compute_variance_decls:
     Env.t ->
-    (Ident.t * type_declaration * type_declaration * class_declaration *
-       cltype_declaration * ((bool * bool) list * Location.t)) list ->
-    (type_declaration * type_declaration * class_declaration *
-       cltype_declaration) list
+    (Ident.t * Types.type_declaration * Types.type_declaration *
+     Types.class_declaration * Types.class_type_declaration *
+     'a Typedtree.class_infos) list ->
+    (Types.type_declaration * Types.type_declaration *
+     Types.class_declaration * Types.class_type_declaration) list
 
 type error =
     Repeated_parameter
@@ -64,18 +69,19 @@ type error =
   | Recursive_abbrev of string
   | Definition_mismatch of type_expr * Includecore.type_mismatch list
   | Constraint_failed of type_expr * type_expr
-  | Unconsistent_constraint of (type_expr * type_expr) list
-  | Type_clash of (type_expr * type_expr) list
+  | Inconsistent_constraint of Env.t * (type_expr * type_expr) list
+  | Type_clash of Env.t * (type_expr * type_expr) list
   | Parameters_differ of Path.t * type_expr * type_expr
   | Null_arity_external
   | Missing_native_external
   | Unbound_type_var of type_expr * type_declaration
   | Unbound_exception of Longident.t
   | Not_an_exception of Longident.t
-  | Bad_variance of int * (bool*bool) * (bool*bool)
+  | Bad_variance of int * (bool*bool*bool) * (bool*bool*bool)
   | Unavailable_type_constructor of Path.t
   | Bad_fixed_type of string
   | Unbound_type_var_exc of type_expr * type_expr
+  | Varying_anonymous
 
 exception Error of Location.t * error
 
